# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/type.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from sarus_data_spec.protobuf import predicate_pb2 as sarus__data__spec_dot_protobuf_dot_predicate__pb2
from sarus_data_spec.protobuf import path_pb2 as sarus__data__spec_dot_protobuf_dot_path__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='sarus_data_spec/protobuf/type.proto',
  package='sarus_data_spec',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n#sarus_data_spec/protobuf/type.proto\x12\x0fsarus_data_spec\x1a(sarus_data_spec/protobuf/predicate.proto\x1a#sarus_data_spec/protobuf/path.proto\"\x86\x14\n\x04Type\x12\x0c\n\x04name\x18\x01 \x01(\t\x12*\n\x04null\x18\x03 \x01(\x0b\x32\x1a.sarus_data_spec.Type.NullH\x00\x12*\n\x04unit\x18\x04 \x01(\x0b\x32\x1a.sarus_data_spec.Type.UnitH\x00\x12\x30\n\x07\x62oolean\x18\x05 \x01(\x0b\x32\x1d.sarus_data_spec.Type.BooleanH\x00\x12\x30\n\x07integer\x18\x06 \x01(\x0b\x32\x1d.sarus_data_spec.Type.IntegerH\x00\x12*\n\x04\x65num\x18\x07 \x01(\x0b\x32\x1a.sarus_data_spec.Type.EnumH\x00\x12,\n\x05\x66loat\x18\x08 \x01(\x0b\x32\x1b.sarus_data_spec.Type.FloatH\x00\x12*\n\x04text\x18\t \x01(\x0b\x32\x1a.sarus_data_spec.Type.TextH\x00\x12,\n\x05\x62ytes\x18\n \x01(\x0b\x32\x1b.sarus_data_spec.Type.BytesH\x00\x12.\n\x06struct\x18\x0b \x01(\x0b\x32\x1c.sarus_data_spec.Type.StructH\x00\x12,\n\x05union\x18\x0c \x01(\x0b\x32\x1b.sarus_data_spec.Type.UnionH\x00\x12\x32\n\x08optional\x18\r \x01(\x0b\x32\x1e.sarus_data_spec.Type.OptionalH\x00\x12*\n\x04list\x18\x0e \x01(\x0b\x32\x1a.sarus_data_spec.Type.ListH\x00\x12,\n\x05\x61rray\x18\x0f \x01(\x0b\x32\x1b.sarus_data_spec.Type.ArrayH\x00\x12\x32\n\x08\x64\x61tetime\x18\x10 \x01(\x0b\x32\x1e.sarus_data_spec.Type.DatetimeH\x00\x12\x38\n\x0b\x63onstrained\x18\x11 \x01(\x0b\x32!.sarus_data_spec.Type.ConstrainedH\x00\x12\x36\n\nhypothesis\x18\x12 \x01(\x0b\x32 .sarus_data_spec.Type.HypothesisH\x00\x12&\n\x02id\x18\x13 \x01(\x0b\x32\x18.sarus_data_spec.Type.IdH\x00\x12\x39\n\nproperties\x18\x02 \x03(\x0b\x32%.sarus_data_spec.Type.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x06\n\x04Null\x1a\x06\n\x04Unit\x1a\t\n\x07\x42oolean\x1a\xa1\x01\n\x07Integer\x12\x30\n\x04\x62\x61se\x18\x01 \x01(\x0e\x32\".sarus_data_spec.Type.Integer.Base\x12\x0b\n\x03min\x18\x02 \x01(\x03\x12\x0b\n\x03max\x18\x03 \x01(\x03\x12\x17\n\x0fpossible_values\x18\x04 \x03(\x03\"1\n\x04\x42\x61se\x12\t\n\x05INT64\x10\x00\x12\t\n\x05INT32\x10\x01\x12\t\n\x05INT16\x10\x02\x12\x08\n\x04INT8\x10\x03\x1a\xde\x01\n\x04\x45num\x12-\n\x04\x62\x61se\x18\x01 \x01(\x0e\x32\x1f.sarus_data_spec.Type.Enum.Base\x12\x0f\n\x07ordered\x18\x02 \x01(\x08\x12\x39\n\x0bname_values\x18\x03 \x03(\x0b\x32$.sarus_data_spec.Type.Enum.NameValue\x1a(\n\tNameValue\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x03\"1\n\x04\x42\x61se\x12\t\n\x05INT64\x10\x00\x12\t\n\x05INT32\x10\x01\x12\t\n\x05INT16\x10\x02\x12\x08\n\x04INT8\x10\x03\x1a\x99\x01\n\x05\x46loat\x12.\n\x04\x62\x61se\x18\x01 \x01(\x0e\x32 .sarus_data_spec.Type.Float.Base\x12\x0b\n\x03min\x18\x02 \x01(\x01\x12\x0b\n\x03max\x18\x03 \x01(\x01\x12\x17\n\x0fpossible_values\x18\x04 \x03(\x02\"-\n\x04\x42\x61se\x12\x0b\n\x07\x46LOAT64\x10\x00\x12\x0b\n\x07\x46LOAT32\x10\x01\x12\x0b\n\x07\x46LOAT16\x10\x02\x1a\x18\n\x04Text\x12\x10\n\x08\x65ncoding\x18\x01 \x01(\t\x1a\x07\n\x05\x42ytes\x1ax\n\x06Struct\x12\x32\n\x06\x66ields\x18\x01 \x03(\x0b\x32\".sarus_data_spec.Type.Struct.Field\x1a:\n\x05\x46ield\x12\x0c\n\x04name\x18\x01 \x01(\t\x12#\n\x04type\x18\x02 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1av\n\x05Union\x12\x31\n\x06\x66ields\x18\x01 \x03(\x0b\x32!.sarus_data_spec.Type.Union.Field\x1a:\n\x05\x46ield\x12\x0c\n\x04name\x18\x01 \x01(\t\x12#\n\x04type\x18\x02 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a/\n\x08Optional\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x1a=\n\x04List\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12\x10\n\x08max_size\x18\x02 \x01(\x03\x1a;\n\x05\x41rray\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12\r\n\x05shape\x18\x02 \x03(\x03\x1a\xb0\x01\n\x08\x44\x61tetime\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x0b\n\x03min\x18\x02 \x01(\t\x12\x0b\n\x03max\x18\x03 \x01(\t\x12\x17\n\x0fpossible_values\x18\x04 \x03(\t\x12\x31\n\x04\x62\x61se\x18\x05 \x01(\x0e\x32#.sarus_data_spec.Type.Datetime.Base\".\n\x04\x42\x61se\x12\x0c\n\x08INT64_NS\x10\x00\x12\x0c\n\x08INT64_MS\x10\x01\x12\n\n\x06STRING\x10\x02\x1a\x62\n\x0b\x43onstrained\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12.\n\nconstraint\x18\x02 \x01(\x0b\x32\x1a.sarus_data_spec.Predicate\x1a\x82\x01\n\nHypothesis\x12\x36\n\x05types\x18\x02 \x03(\x0b\x32\'.sarus_data_spec.Type.Hypothesis.Scored\x1a<\n\x06Scored\x12#\n\x04type\x18\x01 \x01(\x0b\x32\x15.sarus_data_spec.Type\x12\r\n\x05score\x18\x02 \x01(\x01\x1a\xb5\x01\n\x02Id\x12+\n\x04\x62\x61se\x18\x01 \x01(\x0e\x32\x1d.sarus_data_spec.Type.Id.Base\x12\x0e\n\x06unique\x18\x02 \x01(\x08\x12(\n\treference\x18\x03 \x01(\x0b\x32\x15.sarus_data_spec.Path\"H\n\x04\x42\x61se\x12\t\n\x05INT64\x10\x00\x12\t\n\x05INT32\x10\x01\x12\t\n\x05INT16\x10\x02\x12\x08\n\x04INT8\x10\x03\x12\n\n\x06STRING\x10\x04\x12\t\n\x05\x42YTES\x10\x05\x42\x06\n\x04typeb\x06proto3'
  ,
  dependencies=[sarus__data__spec_dot_protobuf_dot_predicate__pb2.DESCRIPTOR,sarus__data__spec_dot_protobuf_dot_path__pb2.DESCRIPTOR,])



_TYPE_INTEGER_BASE = _descriptor.EnumDescriptor(
  name='Base',
  full_name='sarus_data_spec.Type.Integer.Base',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INT64', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INT32', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INT16', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INT8', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1218,
  serialized_end=1267,
)
_sym_db.RegisterEnumDescriptor(_TYPE_INTEGER_BASE)

_TYPE_ENUM_BASE = _descriptor.EnumDescriptor(
  name='Base',
  full_name='sarus_data_spec.Type.Enum.Base',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INT64', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INT32', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INT16', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INT8', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1218,
  serialized_end=1267,
)
_sym_db.RegisterEnumDescriptor(_TYPE_ENUM_BASE)

_TYPE_FLOAT_BASE = _descriptor.EnumDescriptor(
  name='Base',
  full_name='sarus_data_spec.Type.Float.Base',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='FLOAT64', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FLOAT32', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FLOAT16', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1603,
  serialized_end=1648,
)
_sym_db.RegisterEnumDescriptor(_TYPE_FLOAT_BASE)

_TYPE_DATETIME_BASE = _descriptor.EnumDescriptor(
  name='Base',
  full_name='sarus_data_spec.Type.Datetime.Base',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INT64_NS', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INT64_MS', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STRING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2231,
  serialized_end=2277,
)
_sym_db.RegisterEnumDescriptor(_TYPE_DATETIME_BASE)

_TYPE_ID_BASE = _descriptor.EnumDescriptor(
  name='Base',
  full_name='sarus_data_spec.Type.Id.Base',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INT64', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INT32', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INT16', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INT8', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STRING', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BYTES', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2622,
  serialized_end=2694,
)
_sym_db.RegisterEnumDescriptor(_TYPE_ID_BASE)


_TYPE_PROPERTIESENTRY = _descriptor.Descriptor(
  name='PropertiesEntry',
  full_name='sarus_data_spec.Type.PropertiesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='sarus_data_spec.Type.PropertiesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='sarus_data_spec.Type.PropertiesEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1027,
  serialized_end=1076,
)

_TYPE_NULL = _descriptor.Descriptor(
  name='Null',
  full_name='sarus_data_spec.Type.Null',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1078,
  serialized_end=1084,
)

_TYPE_UNIT = _descriptor.Descriptor(
  name='Unit',
  full_name='sarus_data_spec.Type.Unit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1086,
  serialized_end=1092,
)

_TYPE_BOOLEAN = _descriptor.Descriptor(
  name='Boolean',
  full_name='sarus_data_spec.Type.Boolean',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1094,
  serialized_end=1103,
)

_TYPE_INTEGER = _descriptor.Descriptor(
  name='Integer',
  full_name='sarus_data_spec.Type.Integer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='base', full_name='sarus_data_spec.Type.Integer.base', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min', full_name='sarus_data_spec.Type.Integer.min', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max', full_name='sarus_data_spec.Type.Integer.max', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='possible_values', full_name='sarus_data_spec.Type.Integer.possible_values', index=3,
      number=4, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _TYPE_INTEGER_BASE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1106,
  serialized_end=1267,
)

_TYPE_ENUM_NAMEVALUE = _descriptor.Descriptor(
  name='NameValue',
  full_name='sarus_data_spec.Type.Enum.NameValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='sarus_data_spec.Type.Enum.NameValue.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='sarus_data_spec.Type.Enum.NameValue.value', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1401,
  serialized_end=1441,
)

_TYPE_ENUM = _descriptor.Descriptor(
  name='Enum',
  full_name='sarus_data_spec.Type.Enum',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='base', full_name='sarus_data_spec.Type.Enum.base', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ordered', full_name='sarus_data_spec.Type.Enum.ordered', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name_values', full_name='sarus_data_spec.Type.Enum.name_values', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_TYPE_ENUM_NAMEVALUE, ],
  enum_types=[
    _TYPE_ENUM_BASE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1270,
  serialized_end=1492,
)

_TYPE_FLOAT = _descriptor.Descriptor(
  name='Float',
  full_name='sarus_data_spec.Type.Float',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='base', full_name='sarus_data_spec.Type.Float.base', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min', full_name='sarus_data_spec.Type.Float.min', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max', full_name='sarus_data_spec.Type.Float.max', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='possible_values', full_name='sarus_data_spec.Type.Float.possible_values', index=3,
      number=4, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _TYPE_FLOAT_BASE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1495,
  serialized_end=1648,
)

_TYPE_TEXT = _descriptor.Descriptor(
  name='Text',
  full_name='sarus_data_spec.Type.Text',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='encoding', full_name='sarus_data_spec.Type.Text.encoding', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1650,
  serialized_end=1674,
)

_TYPE_BYTES = _descriptor.Descriptor(
  name='Bytes',
  full_name='sarus_data_spec.Type.Bytes',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1676,
  serialized_end=1683,
)

_TYPE_STRUCT_FIELD = _descriptor.Descriptor(
  name='Field',
  full_name='sarus_data_spec.Type.Struct.Field',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='sarus_data_spec.Type.Struct.Field.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='sarus_data_spec.Type.Struct.Field.type', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1747,
  serialized_end=1805,
)

_TYPE_STRUCT = _descriptor.Descriptor(
  name='Struct',
  full_name='sarus_data_spec.Type.Struct',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='fields', full_name='sarus_data_spec.Type.Struct.fields', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_TYPE_STRUCT_FIELD, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1685,
  serialized_end=1805,
)

_TYPE_UNION_FIELD = _descriptor.Descriptor(
  name='Field',
  full_name='sarus_data_spec.Type.Union.Field',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='sarus_data_spec.Type.Union.Field.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='sarus_data_spec.Type.Union.Field.type', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1747,
  serialized_end=1805,
)

_TYPE_UNION = _descriptor.Descriptor(
  name='Union',
  full_name='sarus_data_spec.Type.Union',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='fields', full_name='sarus_data_spec.Type.Union.fields', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_TYPE_UNION_FIELD, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1807,
  serialized_end=1925,
)

_TYPE_OPTIONAL = _descriptor.Descriptor(
  name='Optional',
  full_name='sarus_data_spec.Type.Optional',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='sarus_data_spec.Type.Optional.type', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1927,
  serialized_end=1974,
)

_TYPE_LIST = _descriptor.Descriptor(
  name='List',
  full_name='sarus_data_spec.Type.List',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='sarus_data_spec.Type.List.type', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_size', full_name='sarus_data_spec.Type.List.max_size', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1976,
  serialized_end=2037,
)

_TYPE_ARRAY = _descriptor.Descriptor(
  name='Array',
  full_name='sarus_data_spec.Type.Array',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='sarus_data_spec.Type.Array.type', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='shape', full_name='sarus_data_spec.Type.Array.shape', index=1,
      number=2, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2039,
  serialized_end=2098,
)

_TYPE_DATETIME = _descriptor.Descriptor(
  name='Datetime',
  full_name='sarus_data_spec.Type.Datetime',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='format', full_name='sarus_data_spec.Type.Datetime.format', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min', full_name='sarus_data_spec.Type.Datetime.min', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max', full_name='sarus_data_spec.Type.Datetime.max', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='possible_values', full_name='sarus_data_spec.Type.Datetime.possible_values', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='base', full_name='sarus_data_spec.Type.Datetime.base', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _TYPE_DATETIME_BASE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2101,
  serialized_end=2277,
)

_TYPE_CONSTRAINED = _descriptor.Descriptor(
  name='Constrained',
  full_name='sarus_data_spec.Type.Constrained',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='sarus_data_spec.Type.Constrained.type', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='constraint', full_name='sarus_data_spec.Type.Constrained.constraint', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2279,
  serialized_end=2377,
)

_TYPE_HYPOTHESIS_SCORED = _descriptor.Descriptor(
  name='Scored',
  full_name='sarus_data_spec.Type.Hypothesis.Scored',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='sarus_data_spec.Type.Hypothesis.Scored.type', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='score', full_name='sarus_data_spec.Type.Hypothesis.Scored.score', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2450,
  serialized_end=2510,
)

_TYPE_HYPOTHESIS = _descriptor.Descriptor(
  name='Hypothesis',
  full_name='sarus_data_spec.Type.Hypothesis',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='types', full_name='sarus_data_spec.Type.Hypothesis.types', index=0,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_TYPE_HYPOTHESIS_SCORED, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2380,
  serialized_end=2510,
)

_TYPE_ID = _descriptor.Descriptor(
  name='Id',
  full_name='sarus_data_spec.Type.Id',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='base', full_name='sarus_data_spec.Type.Id.base', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='unique', full_name='sarus_data_spec.Type.Id.unique', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reference', full_name='sarus_data_spec.Type.Id.reference', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _TYPE_ID_BASE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2513,
  serialized_end=2694,
)

_TYPE = _descriptor.Descriptor(
  name='Type',
  full_name='sarus_data_spec.Type',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='sarus_data_spec.Type.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='null', full_name='sarus_data_spec.Type.null', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='unit', full_name='sarus_data_spec.Type.unit', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='boolean', full_name='sarus_data_spec.Type.boolean', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='integer', full_name='sarus_data_spec.Type.integer', index=4,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='enum', full_name='sarus_data_spec.Type.enum', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='float', full_name='sarus_data_spec.Type.float', index=6,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text', full_name='sarus_data_spec.Type.text', index=7,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bytes', full_name='sarus_data_spec.Type.bytes', index=8,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='struct', full_name='sarus_data_spec.Type.struct', index=9,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='union', full_name='sarus_data_spec.Type.union', index=10,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='optional', full_name='sarus_data_spec.Type.optional', index=11,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='list', full_name='sarus_data_spec.Type.list', index=12,
      number=14, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='array', full_name='sarus_data_spec.Type.array', index=13,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='datetime', full_name='sarus_data_spec.Type.datetime', index=14,
      number=16, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='constrained', full_name='sarus_data_spec.Type.constrained', index=15,
      number=17, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='hypothesis', full_name='sarus_data_spec.Type.hypothesis', index=16,
      number=18, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='id', full_name='sarus_data_spec.Type.id', index=17,
      number=19, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='properties', full_name='sarus_data_spec.Type.properties', index=18,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_TYPE_PROPERTIESENTRY, _TYPE_NULL, _TYPE_UNIT, _TYPE_BOOLEAN, _TYPE_INTEGER, _TYPE_ENUM, _TYPE_FLOAT, _TYPE_TEXT, _TYPE_BYTES, _TYPE_STRUCT, _TYPE_UNION, _TYPE_OPTIONAL, _TYPE_LIST, _TYPE_ARRAY, _TYPE_DATETIME, _TYPE_CONSTRAINED, _TYPE_HYPOTHESIS, _TYPE_ID, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='type', full_name='sarus_data_spec.Type.type',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=136,
  serialized_end=2702,
)

_TYPE_PROPERTIESENTRY.containing_type = _TYPE
_TYPE_NULL.containing_type = _TYPE
_TYPE_UNIT.containing_type = _TYPE
_TYPE_BOOLEAN.containing_type = _TYPE
_TYPE_INTEGER.fields_by_name['base'].enum_type = _TYPE_INTEGER_BASE
_TYPE_INTEGER.containing_type = _TYPE
_TYPE_INTEGER_BASE.containing_type = _TYPE_INTEGER
_TYPE_ENUM_NAMEVALUE.containing_type = _TYPE_ENUM
_TYPE_ENUM.fields_by_name['base'].enum_type = _TYPE_ENUM_BASE
_TYPE_ENUM.fields_by_name['name_values'].message_type = _TYPE_ENUM_NAMEVALUE
_TYPE_ENUM.containing_type = _TYPE
_TYPE_ENUM_BASE.containing_type = _TYPE_ENUM
_TYPE_FLOAT.fields_by_name['base'].enum_type = _TYPE_FLOAT_BASE
_TYPE_FLOAT.containing_type = _TYPE
_TYPE_FLOAT_BASE.containing_type = _TYPE_FLOAT
_TYPE_TEXT.containing_type = _TYPE
_TYPE_BYTES.containing_type = _TYPE
_TYPE_STRUCT_FIELD.fields_by_name['type'].message_type = _TYPE
_TYPE_STRUCT_FIELD.containing_type = _TYPE_STRUCT
_TYPE_STRUCT.fields_by_name['fields'].message_type = _TYPE_STRUCT_FIELD
_TYPE_STRUCT.containing_type = _TYPE
_TYPE_UNION_FIELD.fields_by_name['type'].message_type = _TYPE
_TYPE_UNION_FIELD.containing_type = _TYPE_UNION
_TYPE_UNION.fields_by_name['fields'].message_type = _TYPE_UNION_FIELD
_TYPE_UNION.containing_type = _TYPE
_TYPE_OPTIONAL.fields_by_name['type'].message_type = _TYPE
_TYPE_OPTIONAL.containing_type = _TYPE
_TYPE_LIST.fields_by_name['type'].message_type = _TYPE
_TYPE_LIST.containing_type = _TYPE
_TYPE_ARRAY.fields_by_name['type'].message_type = _TYPE
_TYPE_ARRAY.containing_type = _TYPE
_TYPE_DATETIME.fields_by_name['base'].enum_type = _TYPE_DATETIME_BASE
_TYPE_DATETIME.containing_type = _TYPE
_TYPE_DATETIME_BASE.containing_type = _TYPE_DATETIME
_TYPE_CONSTRAINED.fields_by_name['type'].message_type = _TYPE
_TYPE_CONSTRAINED.fields_by_name['constraint'].message_type = sarus__data__spec_dot_protobuf_dot_predicate__pb2._PREDICATE
_TYPE_CONSTRAINED.containing_type = _TYPE
_TYPE_HYPOTHESIS_SCORED.fields_by_name['type'].message_type = _TYPE
_TYPE_HYPOTHESIS_SCORED.containing_type = _TYPE_HYPOTHESIS
_TYPE_HYPOTHESIS.fields_by_name['types'].message_type = _TYPE_HYPOTHESIS_SCORED
_TYPE_HYPOTHESIS.containing_type = _TYPE
_TYPE_ID.fields_by_name['base'].enum_type = _TYPE_ID_BASE
_TYPE_ID.fields_by_name['reference'].message_type = sarus__data__spec_dot_protobuf_dot_path__pb2._PATH
_TYPE_ID.containing_type = _TYPE
_TYPE_ID_BASE.containing_type = _TYPE_ID
_TYPE.fields_by_name['null'].message_type = _TYPE_NULL
_TYPE.fields_by_name['unit'].message_type = _TYPE_UNIT
_TYPE.fields_by_name['boolean'].message_type = _TYPE_BOOLEAN
_TYPE.fields_by_name['integer'].message_type = _TYPE_INTEGER
_TYPE.fields_by_name['enum'].message_type = _TYPE_ENUM
_TYPE.fields_by_name['float'].message_type = _TYPE_FLOAT
_TYPE.fields_by_name['text'].message_type = _TYPE_TEXT
_TYPE.fields_by_name['bytes'].message_type = _TYPE_BYTES
_TYPE.fields_by_name['struct'].message_type = _TYPE_STRUCT
_TYPE.fields_by_name['union'].message_type = _TYPE_UNION
_TYPE.fields_by_name['optional'].message_type = _TYPE_OPTIONAL
_TYPE.fields_by_name['list'].message_type = _TYPE_LIST
_TYPE.fields_by_name['array'].message_type = _TYPE_ARRAY
_TYPE.fields_by_name['datetime'].message_type = _TYPE_DATETIME
_TYPE.fields_by_name['constrained'].message_type = _TYPE_CONSTRAINED
_TYPE.fields_by_name['hypothesis'].message_type = _TYPE_HYPOTHESIS
_TYPE.fields_by_name['id'].message_type = _TYPE_ID
_TYPE.fields_by_name['properties'].message_type = _TYPE_PROPERTIESENTRY
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['null'])
_TYPE.fields_by_name['null'].containing_oneof = _TYPE.oneofs_by_name['type']
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['unit'])
_TYPE.fields_by_name['unit'].containing_oneof = _TYPE.oneofs_by_name['type']
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['boolean'])
_TYPE.fields_by_name['boolean'].containing_oneof = _TYPE.oneofs_by_name['type']
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['integer'])
_TYPE.fields_by_name['integer'].containing_oneof = _TYPE.oneofs_by_name['type']
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['enum'])
_TYPE.fields_by_name['enum'].containing_oneof = _TYPE.oneofs_by_name['type']
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['float'])
_TYPE.fields_by_name['float'].containing_oneof = _TYPE.oneofs_by_name['type']
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['text'])
_TYPE.fields_by_name['text'].containing_oneof = _TYPE.oneofs_by_name['type']
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['bytes'])
_TYPE.fields_by_name['bytes'].containing_oneof = _TYPE.oneofs_by_name['type']
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['struct'])
_TYPE.fields_by_name['struct'].containing_oneof = _TYPE.oneofs_by_name['type']
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['union'])
_TYPE.fields_by_name['union'].containing_oneof = _TYPE.oneofs_by_name['type']
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['optional'])
_TYPE.fields_by_name['optional'].containing_oneof = _TYPE.oneofs_by_name['type']
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['list'])
_TYPE.fields_by_name['list'].containing_oneof = _TYPE.oneofs_by_name['type']
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['array'])
_TYPE.fields_by_name['array'].containing_oneof = _TYPE.oneofs_by_name['type']
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['datetime'])
_TYPE.fields_by_name['datetime'].containing_oneof = _TYPE.oneofs_by_name['type']
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['constrained'])
_TYPE.fields_by_name['constrained'].containing_oneof = _TYPE.oneofs_by_name['type']
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['hypothesis'])
_TYPE.fields_by_name['hypothesis'].containing_oneof = _TYPE.oneofs_by_name['type']
_TYPE.oneofs_by_name['type'].fields.append(
  _TYPE.fields_by_name['id'])
_TYPE.fields_by_name['id'].containing_oneof = _TYPE.oneofs_by_name['type']
DESCRIPTOR.message_types_by_name['Type'] = _TYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Type = _reflection.GeneratedProtocolMessageType('Type', (_message.Message,), {

  'PropertiesEntry' : _reflection.GeneratedProtocolMessageType('PropertiesEntry', (_message.Message,), {
    'DESCRIPTOR' : _TYPE_PROPERTIESENTRY,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.PropertiesEntry)
    })
  ,

  'Null' : _reflection.GeneratedProtocolMessageType('Null', (_message.Message,), {
    'DESCRIPTOR' : _TYPE_NULL,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Null)
    })
  ,

  'Unit' : _reflection.GeneratedProtocolMessageType('Unit', (_message.Message,), {
    'DESCRIPTOR' : _TYPE_UNIT,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Unit)
    })
  ,

  'Boolean' : _reflection.GeneratedProtocolMessageType('Boolean', (_message.Message,), {
    'DESCRIPTOR' : _TYPE_BOOLEAN,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Boolean)
    })
  ,

  'Integer' : _reflection.GeneratedProtocolMessageType('Integer', (_message.Message,), {
    'DESCRIPTOR' : _TYPE_INTEGER,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Integer)
    })
  ,

  'Enum' : _reflection.GeneratedProtocolMessageType('Enum', (_message.Message,), {

    'NameValue' : _reflection.GeneratedProtocolMessageType('NameValue', (_message.Message,), {
      'DESCRIPTOR' : _TYPE_ENUM_NAMEVALUE,
      '__module__' : 'sarus_data_spec.protobuf.type_pb2'
      # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Enum.NameValue)
      })
    ,
    'DESCRIPTOR' : _TYPE_ENUM,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Enum)
    })
  ,

  'Float' : _reflection.GeneratedProtocolMessageType('Float', (_message.Message,), {
    'DESCRIPTOR' : _TYPE_FLOAT,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Float)
    })
  ,

  'Text' : _reflection.GeneratedProtocolMessageType('Text', (_message.Message,), {
    'DESCRIPTOR' : _TYPE_TEXT,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Text)
    })
  ,

  'Bytes' : _reflection.GeneratedProtocolMessageType('Bytes', (_message.Message,), {
    'DESCRIPTOR' : _TYPE_BYTES,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Bytes)
    })
  ,

  'Struct' : _reflection.GeneratedProtocolMessageType('Struct', (_message.Message,), {

    'Field' : _reflection.GeneratedProtocolMessageType('Field', (_message.Message,), {
      'DESCRIPTOR' : _TYPE_STRUCT_FIELD,
      '__module__' : 'sarus_data_spec.protobuf.type_pb2'
      # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Struct.Field)
      })
    ,
    'DESCRIPTOR' : _TYPE_STRUCT,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Struct)
    })
  ,

  'Union' : _reflection.GeneratedProtocolMessageType('Union', (_message.Message,), {

    'Field' : _reflection.GeneratedProtocolMessageType('Field', (_message.Message,), {
      'DESCRIPTOR' : _TYPE_UNION_FIELD,
      '__module__' : 'sarus_data_spec.protobuf.type_pb2'
      # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Union.Field)
      })
    ,
    'DESCRIPTOR' : _TYPE_UNION,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Union)
    })
  ,

  'Optional' : _reflection.GeneratedProtocolMessageType('Optional', (_message.Message,), {
    'DESCRIPTOR' : _TYPE_OPTIONAL,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Optional)
    })
  ,

  'List' : _reflection.GeneratedProtocolMessageType('List', (_message.Message,), {
    'DESCRIPTOR' : _TYPE_LIST,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.List)
    })
  ,

  'Array' : _reflection.GeneratedProtocolMessageType('Array', (_message.Message,), {
    'DESCRIPTOR' : _TYPE_ARRAY,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Array)
    })
  ,

  'Datetime' : _reflection.GeneratedProtocolMessageType('Datetime', (_message.Message,), {
    'DESCRIPTOR' : _TYPE_DATETIME,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Datetime)
    })
  ,

  'Constrained' : _reflection.GeneratedProtocolMessageType('Constrained', (_message.Message,), {
    'DESCRIPTOR' : _TYPE_CONSTRAINED,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Constrained)
    })
  ,

  'Hypothesis' : _reflection.GeneratedProtocolMessageType('Hypothesis', (_message.Message,), {

    'Scored' : _reflection.GeneratedProtocolMessageType('Scored', (_message.Message,), {
      'DESCRIPTOR' : _TYPE_HYPOTHESIS_SCORED,
      '__module__' : 'sarus_data_spec.protobuf.type_pb2'
      # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Hypothesis.Scored)
      })
    ,
    'DESCRIPTOR' : _TYPE_HYPOTHESIS,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Hypothesis)
    })
  ,

  'Id' : _reflection.GeneratedProtocolMessageType('Id', (_message.Message,), {
    'DESCRIPTOR' : _TYPE_ID,
    '__module__' : 'sarus_data_spec.protobuf.type_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Type.Id)
    })
  ,
  'DESCRIPTOR' : _TYPE,
  '__module__' : 'sarus_data_spec.protobuf.type_pb2'
  # @@protoc_insertion_point(class_scope:sarus_data_spec.Type)
  })
_sym_db.RegisterMessage(Type)
_sym_db.RegisterMessage(Type.PropertiesEntry)
_sym_db.RegisterMessage(Type.Null)
_sym_db.RegisterMessage(Type.Unit)
_sym_db.RegisterMessage(Type.Boolean)
_sym_db.RegisterMessage(Type.Integer)
_sym_db.RegisterMessage(Type.Enum)
_sym_db.RegisterMessage(Type.Enum.NameValue)
_sym_db.RegisterMessage(Type.Float)
_sym_db.RegisterMessage(Type.Text)
_sym_db.RegisterMessage(Type.Bytes)
_sym_db.RegisterMessage(Type.Struct)
_sym_db.RegisterMessage(Type.Struct.Field)
_sym_db.RegisterMessage(Type.Union)
_sym_db.RegisterMessage(Type.Union.Field)
_sym_db.RegisterMessage(Type.Optional)
_sym_db.RegisterMessage(Type.List)
_sym_db.RegisterMessage(Type.Array)
_sym_db.RegisterMessage(Type.Datetime)
_sym_db.RegisterMessage(Type.Constrained)
_sym_db.RegisterMessage(Type.Hypothesis)
_sym_db.RegisterMessage(Type.Hypothesis.Scored)
_sym_db.RegisterMessage(Type.Id)


_TYPE_PROPERTIESENTRY._options = None
# @@protoc_insertion_point(module_scope)



from enum import Enum

USER_COLUMN = 'sarus_protected_entity'
WEIGHTS = 'sarus_weights'
PUBLIC = 'sarus_is_public'
LIST_VALUES = 'sarus_list_values'
ARRAY_VALUES = 'sarus_array_values'
OPTIONAL_VALUE = 'sarus_optional_value'
TEXT_MIN_LENGTH = 'min_length'
TEXT_MAX_LENGTH = 'max_length'
TEXT_CHARSET = 'text_char_set'
SQL_MAPPING = 'sql_mapping'
NON_EMPTY_PROTECTED_PATHS = 'non_zero_protected_values'
STRUCT_KIND = 'merge_paths'
TO_MERGE = 'fks_for_merging'
FLOAT_DISTRIBUTION = 'distribution_model'
FLOAT_DIST_PARAMS = 'parameters'
MAX_MAX_MULT = 'max_max_multiplicity'


class StructKind(Enum):
    HAS_PE = '0'
    NO_PE = '1'
    TO_MERGE = '2'


DATETIME_YEAR = 'year'
DATETIME_MONTH = 'month'
DATETIME_DAY = 'day'
DATETIME_HOUR = 'hour'
DATETIME_MINUTES = 'minutes'
DATETIME_SECONDS = 'seconds'


# Statuses constants

TO_PARQUET = 'to_parquet'
CACHE = TO_PARQUET
CACHE_PATH = 'path'



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sarus_data_spec/protobuf/path.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='sarus_data_spec/protobuf/path.proto',
  package='sarus_data_spec',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n#sarus_data_spec/protobuf/path.proto\x12\x0fsarus_data_spec\"\xa9\x01\n\x04Path\x12\r\n\x05label\x18\x01 \x01(\t\x12$\n\x05paths\x18\x02 \x03(\x0b\x32\x15.sarus_data_spec.Path\x12\x39\n\nproperties\x18\x03 \x03(\x0b\x32%.sarus_data_spec.Path.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x62\x06proto3'
)




_PATH_PROPERTIESENTRY = _descriptor.Descriptor(
  name='PropertiesEntry',
  full_name='sarus_data_spec.Path.PropertiesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='sarus_data_spec.Path.PropertiesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='sarus_data_spec.Path.PropertiesEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=177,
  serialized_end=226,
)

_PATH = _descriptor.Descriptor(
  name='Path',
  full_name='sarus_data_spec.Path',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='label', full_name='sarus_data_spec.Path.label', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='paths', full_name='sarus_data_spec.Path.paths', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='properties', full_name='sarus_data_spec.Path.properties', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_PATH_PROPERTIESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=57,
  serialized_end=226,
)

_PATH_PROPERTIESENTRY.containing_type = _PATH
_PATH.fields_by_name['paths'].message_type = _PATH
_PATH.fields_by_name['properties'].message_type = _PATH_PROPERTIESENTRY
DESCRIPTOR.message_types_by_name['Path'] = _PATH
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Path = _reflection.GeneratedProtocolMessageType('Path', (_message.Message,), {

  'PropertiesEntry' : _reflection.GeneratedProtocolMessageType('PropertiesEntry', (_message.Message,), {
    'DESCRIPTOR' : _PATH_PROPERTIESENTRY,
    '__module__' : 'sarus_data_spec.protobuf.path_pb2'
    # @@protoc_insertion_point(class_scope:sarus_data_spec.Path.PropertiesEntry)
    })
  ,
  'DESCRIPTOR' : _PATH,
  '__module__' : 'sarus_data_spec.protobuf.path_pb2'
  # @@protoc_insertion_point(class_scope:sarus_data_spec.Path)
  })
_sym_db.RegisterMessage(Path)
_sym_db.RegisterMessage(Path.PropertiesEntry)


_PATH_PROPERTIESENTRY._options = None
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from typing import (
    Any,
    Collection,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    cast,
)
import pickle as pkl

from sarus_data_spec.base import Referring
from sarus_data_spec.transform import Transform
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


class Scalar(Referring[sp.Scalar]):
    """A python class to describe scalars"""

    def __init__(self, protobuf: sp.Scalar) -> None:
        if protobuf.spec.HasField("transformed"):
            transformed = protobuf.spec.transformed
            self._referred = {
                transformed.transform,
                *transformed.arguments,
                *list(transformed.named_arguments.values()),
            }

        super().__init__(protobuf=protobuf)

    def prototype(self) -> Type[sp.Scalar]:
        """Return the type of the underlying protobuf."""
        return sp.Scalar

    def name(self) -> str:
        return self._protobuf.name

    def doc(self) -> str:
        return self._protobuf.doc

    def is_transformed(self) -> bool:
        """Is the scalar composed."""
        return self._protobuf.spec.HasField("transformed")

    def is_remote(self) -> bool:
        """Is the dataspec a remotely defined dataset."""
        return self.manager().is_remote(self)

    def is_model(self) -> bool:
        """Is the scalar a model."""
        return self._protobuf.spec.HasField("model")

    def status(self) -> st.Status:
        return cast(st.Status, self.manager().status(self))

    def transform(self) -> st.Transform:
        return cast(
            st.Transform,
            self.storage().referrable(
                self.protobuf().spec.transformed.transform
            ),
        )

    def parents(
        self,
    ) -> Tuple[List[st.DataSpec], Dict[str, st.DataSpec]]:
        if not self.is_transformed():
            return list(), dict()

        args_id = self._protobuf.spec.transformed.arguments
        kwargs_id = self._protobuf.spec.transformed.named_arguments

        args_parents = [
            cast(st.DataSpec, self.storage().referrable(uuid))
            for uuid in args_id
        ]
        kwargs_parents = {
            name: cast(st.DataSpec, self.storage().referrable(uuid))
            for name, uuid in kwargs_id.items()
        }

        return args_parents, kwargs_parents

    def is_compliant(
        self,
        kind: st.ConstraintKind,
        public_context: List[str],
        epsilon: Optional[float] = None,
    ) -> bool:
        return self.manager().is_compliant(self, kind, public_context, epsilon)

    def variant(
        self,
        kind: st.ConstraintKind,
        public_context: List[str],
        epsilon: Optional[float] = None,
    ) -> Optional[st.DataSpec]:
        return self.manager().variant(self, kind, public_context, epsilon)

    def variant_constraint(self) -> Optional[st.VariantConstraint]:
        return self.manager().variant_constraint(self)

    def variants(self) -> Collection[st.DataSpec]:
        return self.manager().variants(self)

    def verifies(
        self,
        kind: st.ConstraintKind,
        public_context: Collection[str],
        epsilon: Optional[float],
    ) -> bool:
        variant_constraint = self.variant_constraint()
        if variant_constraint:
            return self.manager().verifies(
                variant_constraint=variant_constraint,
                kind=kind,
                public_context=public_context,
                epsilon=epsilon,
            )
        else:
            return False

    def value(self) -> st.DataSpecValue:
        return self.manager().value(self)

    async def async_value(self) -> st.DataSpecValue:
        return await self.manager().async_value(self)

    # A Visitor acceptor
    def accept(self, visitor: st.Visitor) -> None:
        visitor.all(self)
        if self.is_transformed():
            visitor.transformed(
                self,
                cast(
                    Transform,
                    self.storage().referrable(
                        self._protobuf.spec.transformed.transform
                    ),
                ),
                *(
                    cast(Scalar, self.storage().referrable(arg))
                    for arg in self._protobuf.spec.transformed.arguments
                ),
                **{
                    name: cast(Scalar, self.storage().referrable(arg))
                    for name, arg in self._protobuf.spec.transformed.named_arguments.items()  # noqa: E501
                },
            )
        else:
            visitor.other(self)

    def dot(self) -> str:
        """return a graphviz representation of the scalar"""

        class Dot(st.Visitor):
            visited: Set[st.DataSpec] = set()
            nodes: Dict[str, Tuple[str, str]] = {}
            edges: Dict[Tuple[str, str], str] = {}

            def transformed(
                self,
                visited: st.DataSpec,
                transform: st.Transform,
                *arguments: st.DataSpec,
                **named_arguments: st.DataSpec,
            ) -> None:
                if visited not in self.visited:
                    if visited.prototype() == sp.Dataset:
                        self.nodes[visited.uuid()] = (
                            visited.name(),
                            "Dataset",
                        )
                    else:
                        self.nodes[visited.uuid()] = (visited.name(), "Scalar")

                    if not visited.is_remote():
                        for argument in arguments:
                            self.edges[
                                (argument.uuid(), visited.uuid())
                            ] = transform.name()
                            argument.accept(self)
                        for _, argument in named_arguments.items():
                            self.edges[
                                (argument.uuid(), visited.uuid())
                            ] = transform.name()
                            argument.accept(self)
                    self.visited.add(visited)

            def other(self, visited: st.DataSpec) -> None:
                if visited.prototype() == sp.Dataset:
                    self.nodes[visited.uuid()] = (
                        visited.name(),
                        "Dataset",
                    )
                else:
                    self.nodes[visited.uuid()] = (visited.name(), "Scalar")

        visitor = Dot()
        self.accept(visitor)
        result = 'digraph {'
        for uuid, (label, node_type) in visitor.nodes.items():
            shape = "polygon" if node_type == "Scalar" else "ellipse"
            result += (
                f'\n"{uuid}" [label="{label} ({uuid[:2]})", shape={shape}];'
            )
        for (u1, u2), label in visitor.edges.items():
            result += f'\n"{u1}" -> "{u2}" [label="{label} ({uuid[:2]})"];'
        result += '}'
        return result


# Builders
def model(
    model_class: sp.Scalar.Model.ModelClass.V, *args: Any, **kwargs: Any
) -> Scalar:
    return Scalar(
        sp.Scalar(
            name=sp.Scalar.Model.ModelClass.Name(model_class),
            spec=sp.Scalar.Spec(
                model=sp.Scalar.Model(
                    model_class=model_class,
                    arguments=pkl.dumps(args),
                    named_arguments=pkl.dumps(kwargs),
                )
            ),
        )
    )


class Visitor:
    """A visitor class for Scalar"""

    def all(self, visited: Scalar) -> None:
        return

    def transformed(
        self,
        visited: Scalar,
        transform: Transform,
        *arguments: Scalar,
        **named_arguments: Scalar,
    ) -> None:
        return

    def other(self, visited: Scalar) -> None:
        return



from typing import Final

from sarus_data_spec.attribute import Attribute
from sarus_data_spec.context import push_global_context
from sarus_data_spec.dataset import Dataset
from sarus_data_spec.scalar import Scalar
from sarus_data_spec.status import Status
from sarus_data_spec.transform import Transform
from sarus_data_spec.variant_constraint import VariantConstraint

"""A library to manage Sarus datasets"""
# pylint: disable=unused-variable

PACKAGE_NAME: Final[str] = 'sarus_data_spec'
VERSION: Final[str] = '1.3.6'

try:
    # The local context is absent from the public release
    import sarus_data_spec.context.local as sc

    push_global_context(sc.Local())
except ModuleNotFoundError:
    pass



from typing import Any, AsyncIterator

import pyarrow as pa

import sarus_data_spec.typing as st

from .external.routing import arrow_external, external
from .standard.sample import arrow_sample


async def transformed_dataset_arrow(
    dataset: st.Dataset, batch_size: int
) -> AsyncIterator[pa.RecordBatch]:
    """Routes a transformed Dataspec to its Arrow implementation."""
    transform = dataset.transform()
    if transform.is_external():
        return await arrow_external(dataset, batch_size)
    elif transform.protobuf().spec.HasField('sample'):
        return await arrow_sample(dataset, batch_size)
    else:
        raise NotImplementedError(
            f"{transform.protobuf().spec.WhichOneof('spec')}"
        )


async def transformed_scalar(scalar: st.Scalar) -> Any:
    """Routes a transformed Scalar to its implementation.

    For now, transformed Scalars are only obtained from external ops. By
    external ops, we mean functions defined in external libraries.
    """
    if scalar.transform().is_external():
        return await external(scalar)
    else:
        raise NotImplementedError(
            f"scalar_transformed for {scalar.transform()}"
        )



"""Protocols describing common object behaviors.
"""
from __future__ import annotations

import typing as t

import pandas as pd
import pyarrow as pa

try:
    import tensorflow as tf
except ModuleNotFoundError:
    print('tensorflow not found, tensorflow datasets not available')

try:
    import sklearn  # noqa: F401
except ModuleNotFoundError:
    print('sklearn not found, sklearn models not available')

try:
    import pandas_profiling  # noqa: F401
except ModuleNotFoundError:
    print('pandas_profiling not found, ProfileReport not available')

from enum import Enum
import datetime as dt

import numpy as np

from sarus_data_spec.protobuf.typing import Protobuf, ProtobufWithUUID
import sarus_data_spec.manager.typing as manager_typing
import sarus_data_spec.protobuf as sp
import sarus_data_spec.storage.typing as storage_typing

if t.TYPE_CHECKING:
    from sklearn import svm

    DataSpecValue = t.Union[pd.DataFrame, np.ndarray, svm.SVC]
else:
    DataSpecValue = t.Any
P = t.TypeVar('P', bound=Protobuf, covariant=True)


@t.runtime_checkable
class HasProtobuf(t.Protocol[P]):
    """An object backed by a protocol buffer message."""

    def protobuf(self) -> P:
        """Returns the underlying protobuf object."""
        ...

    def prototype(self) -> t.Type[P]:
        """Returns the type of protobuf."""
        ...

    def type_name(self) -> str:
        """Returns the name of the type."""
        ...

    def __getitem__(self, key: str) -> str:
        """Returns the property referred by key"""
        ...

    def properties(self) -> t.Mapping[str, str]:
        """Returns the properties"""
        ...


@t.runtime_checkable
class Value(t.Protocol):
    """An object with value semantics."""

    def __bytes__(self) -> bytes:
        ...

    def __repr__(self) -> str:
        ...

    def __str__(self) -> str:
        ...

    def __eq__(self, value: object) -> bool:
        ...

    def __hash__(self) -> int:
        ...


@t.runtime_checkable
class Frozen(t.Protocol):
    """An immutable object."""

    def _freeze(self) -> None:
        """Freeze the state of the object"""
        ...

    def _frozen(self) -> bool:
        """Check if the frozen object was left unchanged"""
        ...


PU = t.TypeVar('PU', bound=ProtobufWithUUID, covariant=True)


@t.runtime_checkable
class Referrable(HasProtobuf[PU], Frozen, t.Protocol[PU]):
    """Can be referred to by uuid."""

    def uuid(self) -> str:
        """Reference to use to refer to this object."""
        ...

    def referring(
        self, type_name: t.Optional[str] = None
    ) -> t.Collection[Referring[ProtobufWithUUID]]:
        """Referring objects pointing to this one."""
        ...

    def storage(self) -> storage_typing.Storage:
        ...

    def manager(self) -> manager_typing.Manager:
        ...


@t.runtime_checkable
class Referring(Referrable[PU], Frozen, t.Protocol[PU]):
    """Is referring to other Referrables"""

    def referred(self) -> t.Collection[Referrable[ProtobufWithUUID]]:
        """Referred by this object."""
        ...


FM = t.TypeVar('FM', bound=Protobuf)


@t.runtime_checkable
class Factory(t.Protocol):
    """Can produce objects from protobuf messages"""

    def register(self, name: str, type: t.Type[HasProtobuf[Protobuf]]) -> None:
        """Registers a class"""
        ...

    def create(self, message: FM) -> HasProtobuf[FM]:
        """Returns a wrapped protobuf"""
        ...


class VariantConstraint(Referrable[sp.VariantConstraint]):
    def constraint_kind(self) -> ConstraintKind:
        ...

    def required_context(self) -> t.List[str]:
        ...

    def epsilon(self) -> t.Optional[float]:
        ...

    def accept(self, visitor: TransformVisitor) -> None:
        ...


# Type alias
DS = t.TypeVar('DS', bound=t.Union[sp.Scalar, sp.Dataset])


class Attribute(Referring[sp.Attribute]):
    def prototype(self) -> t.Type[sp.Attribute]:
        ...


@t.runtime_checkable
class DataSpec(Referring[DS], t.Protocol):
    def parents(
        self,
    ) -> t.Tuple[t.List[DataSpec[DS]], t.Dict[str, DataSpec[DS]]]:
        ...

    def is_compliant(
        self,
        kind: ConstraintKind,
        public_context: t.List[str],
        epsilon: t.Optional[float] = None,
    ) -> bool:
        ...

    def variant(
        self,
        kind: ConstraintKind,
        public_context: t.List[str],
        epsilon: t.Optional[float] = None,
    ) -> t.Optional[DataSpec[DS]]:
        ...

    def variant_constraint(self) -> t.Optional[VariantConstraint]:
        ...

    def verifies(
        self,
        kind: ConstraintKind,
        public_context: t.Collection[str],
        epsilon: t.Optional[float],
    ) -> bool:
        ...

    def variants(self) -> t.Collection[DataSpec]:
        ...

    def name(self) -> str:
        ...

    def doc(self) -> str:
        ...

    def is_transformed(self) -> bool:
        """Is the dataset composed."""
        ...

    def is_remote(self) -> bool:
        """Is the dataspec a remotely defined dataset."""
        ...

    def transform(self) -> Transform:
        ...

    def status(self) -> Status:
        ...

    def accept(self, visitor: Visitor) -> None:
        ...


class Dataset(DataSpec[sp.Dataset], t.Protocol):
    def prototype(self) -> t.Type[sp.Dataset]:
        ...

    def is_synthetic(self) -> bool:
        ...

    def is_protected(self) -> bool:
        ...

    def is_source(self) -> bool:
        ...

    def sources(self) -> t.Set[Dataset]:
        ...

    def schema(self) -> Schema:
        ...

    def size(self) -> Size:
        ...

    def bounds(self) -> Bounds:
        ...

    def marginals(self) -> Marginals:
        ...

    def to_arrow(self, batch_size: int = 10000) -> t.Iterator[pa.RecordBatch]:
        ...

    async def async_to_arrow(
        self, batch_size: int = 10000
    ) -> t.AsyncIterator[pa.RecordBatch]:
        ...

    def spec(self) -> str:
        ...

    def __iter__(self) -> t.Iterator[pa.RecordBatch]:
        ...

    def to_pandas(self) -> pd.DataFrame:
        ...

    async def async_to_pandas(self) -> pd.DataFrame:
        ...

    def to_tensorflow(self) -> tf.data.Dataset:
        ...

    async def async_to_tensorflow(self) -> tf.data.Dataset:
        ...

    def batch(self, batch_size: int) -> Dataset:
        """sets the batch size of the dataset"""
        ...

    def dot(self) -> str:
        """return a graphviz representation of the dataset"""
        ...

    def sql(self, query: str) -> t.List[t.Dict[str, t.Any]]:
        """Executes the sql method on the dataset"""
        ...

    def foreign_keys(self) -> t.Dict[Path, Path]:
        """returns foreign keys of the dataset"""
        ...

    def primary_keys(self) -> t.List[Path]:
        """Returns a list of the paths to all primary keys"""
        ...


class Scalar(DataSpec[sp.Scalar], t.Protocol):
    def prototype(self) -> t.Type[sp.Scalar]:
        """Return the type of the underlying protobuf."""
        ...

    def is_model(self) -> bool:
        ...

    def value(self) -> DataSpecValue:
        ...

    async def async_value(self) -> DataSpecValue:
        ...


class Visitor(t.Protocol):
    """A visitor class for Dataset"""

    def all(self, visited: DataSpec) -> None:
        ...

    def transformed(
        self,
        visited: DataSpec,
        transform: Transform,
        *arguments: DataSpec,
        **named_arguments: DataSpec,
    ) -> None:
        ...

    def other(self, visited: DataSpec) -> None:
        ...


class Bounds(Referring[sp.Bounds], t.Protocol):
    """A python abstract class to describe bounds"""

    def prototype(self) -> t.Type[sp.Bounds]:
        """Return the type of the underlying protobuf."""
        ...

    def dataset(self) -> Dataset:
        ...

    def statistics(self) -> Statistics:
        ...


class Marginals(Referring[sp.Marginals], t.Protocol):
    """A python abstract class to describe marginals"""

    def prototype(self) -> t.Type[sp.Marginals]:
        """Return the type of the underlying protobuf."""
        ...

    def dataset(self) -> Dataset:
        ...

    def statistics(self) -> Statistics:
        ...


class Size(Referring[sp.Size], t.Protocol):
    """A python abstract class to describe size"""

    def prototype(self) -> t.Type[sp.Size]:
        """Return the type of the underlying protobuf."""
        ...

    def dataset(self) -> Dataset:
        ...

    def statistics(self) -> Statistics:
        ...


class Schema(Referring[sp.Schema], t.Protocol):
    """A python abstract class to describe schemas"""

    def prototype(self) -> t.Type[sp.Schema]:
        """Return the type of the underlying protobuf."""
        ...

    def dataset(self) -> Dataset:
        ...

    def to_arrow(self) -> pa.Schema:
        ...

    def type(self) -> Type:
        ...

    def tables(self) -> t.List[Path]:
        ...

    def protected_path(self) -> Path:
        ...

    def data_type(self) -> Type:
        ...


class Status(Referring[sp.Status], t.Protocol):
    """A python abstract class to describe status"""

    def prototype(self) -> t.Type[sp.Status]:
        """Return the type of the underlying protobuf."""
        ...

    def dataset(self) -> Dataset:
        ...

    def datetime(self) -> dt.datetime:
        ...

    def update(
        self,
        task_stages: t.Optional[t.Mapping[str, Stage]],
        properties: t.Optional[t.Mapping[str, str]],
    ) -> Status:
        ...

    def task(self, task: str) -> t.Optional[Stage]:
        ...

    def pending(self) -> bool:
        ...

    def processing(self) -> bool:
        ...

    def ready(self) -> bool:
        ...

    def error(self) -> bool:
        ...

    def owner(
        self,
    ) -> manager_typing.Manager:  # TODO: Maybe find a better name, but this was shadowing the actual manager of this object.  # noqa: E501
        ...


class Stage(HasProtobuf[sp.Status.Stage], t.Protocol):
    def accept(self, visitor: StageVisitor) -> None:
        ...

    def stage(self) -> str:
        ...

    def ready(self) -> bool:
        ...

    def processing(self) -> bool:
        ...

    def pending(self) -> bool:
        ...

    def error(self) -> bool:
        ...


class StageVisitor(t.Protocol):
    """A visitor class for Status/Stage"""

    def pending(self) -> None:
        ...

    def processing(self) -> None:
        ...

    def ready(self) -> None:
        ...

    def error(self) -> None:
        ...


class Transform(Referrable[sp.Transform], t.Protocol):
    """A python abstract class to describe transforms"""

    def prototype(self) -> t.Type[sp.Transform]:
        """Return the type of the underlying protobuf."""
        ...

    def name(self) -> str:
        ...

    def doc(self) -> str:
        ...

    def spec(self) -> str:
        ...

    def is_composed(self) -> bool:
        """Is the transform composed."""
        ...

    def is_variable(self) -> bool:
        """Is the transform a variable."""
        ...

    def is_external(self) -> bool:
        """Is the transform an external operation."""
        ...

    def infer_output_type(
        self, *arguments: DataSpec, **named_arguments: DataSpec
    ) -> t.Tuple[str, t.Callable[[DataSpec], None]]:
        """Guess if the external transform output is a Dataset or a Scalar.

        Registers schema if it is a Dataset and returns the value type.
        """
        ...

    def transforms(self) -> t.Set[Transform]:
        """return all transforms (and avoid infinite recursions/loops)"""
        ...

    def variables(self) -> t.Set[Transform]:
        """Return all the variables from a composed transform"""
        ...

    def compose(
        self,
        *compose_arguments: Transform,
        **compose_named_arguments: Transform,
    ) -> Transform:
        ...

    def apply(
        self,
        *apply_arguments: DataSpec,
        **apply_named_arguments: DataSpec,
    ) -> DataSpec:
        ...

    def abstract(
        self,
        *arguments: t.Union[int, str],
        **named_arguments: t.Union[int, str],
    ) -> Transform:
        ...

    def __call__(
        self,
        *arguments: t.Union[Transform, DataSpec, int, str],
        **named_arguments: t.Union[Transform, DataSpec, int, str],
    ) -> t.Union[Transform, DataSpec]:
        """Applies the transform to another element"""
        ...

    def __mul__(self, argument: Transform) -> Transform:
        ...

    def accept(self, visitor: TransformVisitor) -> None:
        ...


class TransformVisitor(t.Protocol):
    """A visitor class for Transform"""

    def all(self, visited: Transform) -> None:
        ...

    def composed(
        self,
        visited: Transform,
        transform: Transform,
        *arguments: Transform,
        **named_arguments: Transform,
    ) -> None:
        ...

    def variable(
        self,
        visited: Transform,
        position_name: t.Union[int, str] = 0,
    ) -> None:
        ...

    def other(self, visited: Transform) -> None:
        ...


class Path(HasProtobuf[sp.Path], Frozen, Value, t.Protocol):
    """A python class to describe Paths"""

    def prototype(self) -> t.Type[sp.Path]:
        """Return the type of the underlying protobuf."""
        ...

    def to_strings_list(self) -> t.List[t.List[str]]:
        ...

    def to_dict(self) -> t.Dict[str, str]:
        ...

    def label(self) -> str:
        ...

    def sub_paths(self) -> t.List[Path]:
        ...


class Type(HasProtobuf[sp.Type], Frozen, Value, t.Protocol):
    def prototype(self) -> t.Type[sp.Type]:
        """Return the type of the underlying protobuf."""
        ...

    def name(self) -> str:
        """Returns the name of the underlying protobuf."""
        ...

    def latex(self: Type, parenthesized: bool = False) -> str:
        """return a latex representation of the type"""
        ...

    def compact(self: Type, parenthesized: bool = False) -> str:
        """return a compact representation of the type"""
        ...

    def structs(self: Type) -> t.Optional[t.List[Path]]:
        """Returns the path to the first level structs encountered in the
        type.
        For example, Union[Struct1,Union[Struct2[Struct3]] will return only a
        path that brings to Struct1 and Struct2.
        """
        ...

    def get(self, item: Path) -> Type:
        """Return a subtype of the considered type defined by the path."""
        ...

    def leaves(self) -> t.List[Type]:
        """Returns the leaves contained in the type tree structure"""
        ...

    def children(self) -> t.Dict[str, Type]:
        """Returns the children contained in the type tree structure"""
        ...

    # A Visitor acceptor
    def accept(self, visitor: TypeVisitor) -> None:
        ...

    def sub_types(self: Type, item: Path) -> t.List[Type]:
        """Returns the terminal nodes contained in the path"""
        ...

    def default(self: Type) -> pa.Array:
        """Returns an example of arrow array matching the type.
        For an optional type, it sets the default missing value.
        """

    def numpy_default(self: Type) -> np.ndarray:
        """Returns an example of numpy array matching the type.
        For an optional type, it sets the default missing value
        """

    def tensorflow_default(self, is_optional: bool = False) -> t.Any:
        """This methods returns a dictionary with tensors as leaves
        that match the type. For an optional type, we consider
        the case where the field is missing, and set the default value
        for each missing type.
        """

    def example(self: Type) -> pa.Array:
        """Returns an example of arrow array matching the type.
        For an optional type, it returns a non missing
        value of the type.
        """

    def numpy_example(self: Type) -> np.ndarray:
        """Returns an example of numpy array matching the type.
        For an optional type, it returns a non
        missing value of the type.
        """

    def tensorflow_example(self: Type) -> t.Any:
        """Returns an example of a dictionary with tensors as leaves
        that match the type..
        For an optional type, it returns a non missing value of the type.
        """


class IdBase(Enum):
    INT64 = sp.Type.Id.INT64
    INT32 = sp.Type.Id.INT32
    INT16 = sp.Type.Id.INT16
    INT8 = sp.Type.Id.INT8
    STRING = sp.Type.Id.STRING
    BYTES = sp.Type.Id.BYTES


class DatetimeBase(Enum):
    INT64_NS = sp.Type.Datetime.INT64_NS
    INT64_MS = sp.Type.Datetime.INT64_MS
    STRING = sp.Type.Datetime.STRING


class IntegerBase(Enum):
    INT64 = sp.Type.Integer.INT64
    INT32 = sp.Type.Integer.INT32
    INT16 = sp.Type.Integer.INT16
    INT8 = sp.Type.Integer.INT8


class FloatBase(Enum):
    FLOAT64 = sp.Type.Float.FLOAT64
    FLOAT32 = sp.Type.Float.FLOAT32
    FLOAT16 = sp.Type.Float.FLOAT16


class ConstraintKind(Enum):
    SYNTHETIC = sp.ConstraintKind.SYNTHETIC
    PEP = sp.ConstraintKind.PEP
    DP = sp.ConstraintKind.DP
    PUBLIC = sp.ConstraintKind.PUBLIC


class TypeVisitor(t.Protocol):
    """A visitor class for Type"""

    def Null(self) -> None:
        ...

    def Unit(self) -> None:
        ...

    def Boolean(self) -> None:
        ...

    def Id(
        self,
        unique: bool,
        reference: t.Optional[Path] = None,
        base: t.Optional[IdBase] = None,
    ) -> None:
        ...

    def Integer(self, min: int, max: int, base: IntegerBase) -> None:
        ...

    def Enum(
        self,
        name: str,
        name_values: t.Sequence[t.Tuple[str, int]],
        ordered: bool,
    ) -> None:
        ...

    def Float(self, min: float, max: float, base: FloatBase) -> None:
        ...

    def Text(self, encoding: str) -> None:
        ...

    def Bytes(self) -> None:
        ...

    def Struct(
        self, fields: t.Mapping[str, Type], name: t.Optional[str] = None
    ) -> None:
        ...

    def Union(
        self, fields: t.Mapping[str, Type], name: t.Optional[str] = None
    ) -> None:
        ...

    def Optional(self, type: Type, name: t.Optional[str] = None) -> None:
        ...

    def List(
        self, type: Type, max_size: int, name: t.Optional[str] = None
    ) -> None:
        ...

    def Array(
        self,
        type: Type,
        shape: t.Tuple[int, ...],
        name: t.Optional[str] = None,
    ) -> None:
        ...

    def Datetime(
        self, format: str, min: str, max: str, base: DatetimeBase
    ) -> None:
        ...

    def Constrained(
        self,
        type: Type,
        constraint: Predicate,
        name: t.Optional[str] = None,
    ) -> None:
        ...

    def Hypothesis(
        self, *types: t.Tuple[Type, float], name: t.Optional[str] = None
    ) -> None:
        ...


class Predicate(HasProtobuf[sp.Predicate], Frozen, Value, t.Protocol):
    """A python class to describe types"""

    def prototype(self) -> t.Type[sp.Predicate]:
        """Return the type of the underlying protobuf."""

    # A bunch of operators
    def __or__(self, predicate: Predicate) -> Predicate:
        """Union operator"""

    def __and__(self, predicate: Predicate) -> Predicate:
        """Inter operator"""

    def __invert__(self) -> Predicate:
        """Complement"""


class Statistics(HasProtobuf[sp.Statistics], Frozen, Value, t.Protocol):
    """A python class to describe statistics"""

    def prototype(self) -> t.Type[sp.Statistics]:
        """Return the type of the underlying protobuf."""
        ...

    def name(self) -> str:
        ...

    def distribution(self) -> Distribution:
        ...

    def size(self) -> int:
        ...

    def multiplicity(self) -> float:
        ...

    def accept(self, visitor: StatisticsVisitor) -> None:
        ...

    def nodes_statistics(self, path: Path) -> t.List[Statistics]:
        """Returns the List of each statistics corresponding at the leaves
        of path"""
        ...

    def children(self) -> t.Dict[str, Statistics]:
        """Returns the children contained in the type tree structure"""
        ...


class Distribution(HasProtobuf[sp.Distribution], Frozen, Value, t.Protocol):
    """A python class to describe distributions"""

    def prototype(self) -> t.Type[sp.Distribution]:
        """Return the type of the underlying protobuf."""
        ...

    def values(self) -> t.Union[t.List[float], t.List[int]]:
        ...

    def probabilities(self) -> t.List[float]:
        ...

    def names(self) -> t.Union[t.List[bool], t.List[str]]:
        ...

    def min_value(self) -> t.Union[int, float]:
        ...

    def max_value(self) -> t.Union[int, float]:
        ...


class StatisticsVisitor(t.Protocol):
    """A visitor class for Statistics"""

    def Null(self, size: int, multiplicity: float) -> None:
        return

    def Unit(self, size: int, multiplicity: float) -> None:
        return

    def Boolean(
        self,
        size: int,
        multiplicity: float,
        probabilities: t.Optional[t.List[float]] = None,
        names: t.Optional[t.List[bool]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None:
        return

    def Id(self, size: int, multiplicity: float) -> None:
        return

    def Integer(
        self,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None:
        return

    def Enum(
        self,
        size: int,
        multiplicity: float,
        probabilities: t.Optional[t.List[float]] = None,
        names: t.Optional[t.List[str]] = None,
        values: t.Optional[t.List[float]] = None,
        name: str = 'Enum',
    ) -> None:
        return

    def Float(
        self,
        size: int,
        multiplicity: float,
        min_value: float,
        max_value: float,
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[float]] = None,
    ) -> None:
        return

    def Text(
        self,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        example: str = '',
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None:
        return

    def Bytes(self, size: int, multiplicity: float) -> None:
        return

    def Struct(
        self,
        fields: t.Mapping[str, Statistics],
        size: int,
        multiplicity: float,
        name: t.Optional[str] = None,
    ) -> None:
        return

    def Union(
        self,
        fields: t.Mapping[str, Statistics],
        size: int,
        multiplicity: float,
        name: t.Optional[str] = None,
    ) -> None:
        return

    def Optional(
        self, statistics: Statistics, size: int, multiplicity: float
    ) -> None:
        return

    def List(
        self,
        statistics: Statistics,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        name: str = 'List',
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None:
        return

    def Array(
        self,
        statistics: Statistics,
        size: int,
        multiplicity: float,
        min_values: t.Optional[t.List[float]] = None,
        max_values: t.Optional[t.List[float]] = None,
        name: str = 'Array',
        probabilities: t.Optional[t.List[t.List[float]]] = None,
        values: t.Optional[t.List[t.List[float]]] = None,
    ) -> None:
        return

    def Datetime(
        self,
        size: int,
        multiplicity: float,
        min_value: int,
        max_value: int,
        probabilities: t.Optional[t.List[float]] = None,
        values: t.Optional[t.List[int]] = None,
    ) -> None:
        return

    def Constrained(
        self, statistics: Statistics, size: int, multiplicity: float
    ) -> None:
        return


class InferredDistribution(t.Protocol):
    """A python class to to infer user input distribution

    Attributes:
        nparams: number of parameters
    """

    nparams: int

    def estimate_params(self, x: np.ndarray) -> None:
        """estimate distribution parameters (non-DP) from data column"""
        ...

    def log_likelihood(self, x: np.ndarray) -> float:
        """compute log-likelihood of the distribution on data column"""
        ...

    def preprocess(self, x: np.ndarray) -> np.ndarray:
        """Shift/scale data to be able to estimate distribution parameters"""
        ...

    def params(self) -> t.Mapping[str, float]:
        """return distribution parameters"""
        ...


class InferredAlphabet(t.Protocol):
    """A python class to to infer user input charset

    Attributes:
        charset (t.List[int]): list with int representation of unique chars
        complexity (int): charset intervals used to generate the alphabet
            e.g. if alphabet (ascii) is [1,2,3, ..., 126] has complexity=1
            if alphabet is [1,2,3] U [10] = [1,2,3,10] has complexity=2
    """

    charset: t.List[int]
    complexity: int



from __future__ import annotations

from typing import AsyncIterator, Iterator
import asyncio
import logging
import os
import typing

import pandas as pd
import pyarrow as pa

from sarus_data_spec.arrow.array import convert_record_batch
from sarus_data_spec.manager.base import Base
from sarus_data_spec.storage.typing import Storage
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st

try:
    import tensorflow as tf

    from sarus_data_spec.manager.ops.tensorflow.features import (
        deserialize,
        flatten,
        nest,
        serialize,
        to_internal_signature,
    )
    from sarus_data_spec.manager.ops.tensorflow.tensorflow_visitor import (
        convert_tensorflow,
    )
except ModuleNotFoundError:
    pass  # error message printed from typing.py

from .utils import iter_over_async

logger = logging.getLogger(__name__)

BATCH_SIZE = 10000


class BaseAsyncManager(Base):
    """Asynchronous Manager Base implementation.

    Make synchronous methods rely on asynchronous ones for consistency.
    """

    def __init__(self, storage: Storage, protobuf: sp.Manager):
        super().__init__(
            storage=storage,
            protobuf=protobuf,
        )

    async def async_value(self, scalar: st.Scalar) -> st.DataSpecValue:
        raise NotImplementedError("async_value")

    def value(self, scalar: st.Scalar) -> st.DataSpecValue:
        return asyncio.run(self.async_value(scalar=scalar))

    async def async_to_arrow(
        self, dataset: st.Dataset, batch_size: int
    ) -> AsyncIterator[pa.RecordBatch]:
        raise NotImplementedError("async_to_arrow")

    def to_arrow(
        self, dataset: st.Dataset, batch_size: int
    ) -> Iterator[pa.RecordBatch]:
        batches_async_iterator = asyncio.run(
            self.async_to_arrow(dataset=dataset, batch_size=batch_size)
        )
        return iter_over_async(batches_async_iterator)

    async def async_to_pandas(self, dataset: st.Dataset) -> pd.DataFrame:
        batches_async_it = await self.async_to_arrow(
            dataset=dataset, batch_size=BATCH_SIZE
        )
        arrow_batches = [batch async for batch in batches_async_it]
        return pa.Table.from_batches(arrow_batches).to_pandas()

    def to_pandas(self, dataset: st.Dataset) -> pd.DataFrame:
        return asyncio.run(self.async_to_pandas(dataset=dataset))

    def to_tensorflow(self, dataset: st.Dataset) -> tf.data.Dataset:
        return asyncio.run(self.async_to_tensorflow(dataset=dataset))

    async def async_to_tensorflow(
        self, dataset: st.Dataset
    ) -> tf.data.Dataset:

        root_dir = os.path.join(self.parquet_dir, "tfrecords", dataset.uuid())
        schema_type = dataset.schema().type()
        signature = to_internal_signature(schema_type)

        if not os.path.exists(root_dir):
            # the dataset is cached first
            os.makedirs(root_dir)

            flattener = flatten(signature)
            serializer = serialize(signature)
            i = 0
            batches_async_it = await self.async_to_arrow(
                dataset=dataset, batch_size=BATCH_SIZE
            )
            async for batch in batches_async_it:
                filename = os.path.join(root_dir, f"batch_{i}.tfrecord")
                i += 1
                await write_tf_batch(
                    filename, batch, schema_type, flattener, serializer
                )

        # reading from cache
        glob = os.path.join(root_dir, "*.tfrecord")
        filenames = tf.data.Dataset.list_files(glob, shuffle=False)
        deserializer = deserialize(signature)
        nester = nest(signature)
        return tf.data.TFRecordDataset(filenames).map(deserializer).map(nester)


#
async def write_tf_batch(
    filename: str,
    batch: pa.RecordBatch,
    schema_type: st.Type,
    flattener: typing.Callable,
    serializer: typing.Callable,
) -> None:
    with tf.io.TFRecordWriter(filename) as writer:
        batch = convert_tensorflow(
            convert_record_batch(record_batch=batch, _type=schema_type),
            schema_type,
        )
        batch = tf.data.Dataset.from_tensor_slices(batch).map(flattener)
        for row in batch:
            as_bytes = serializer(row)
            writer.write(as_bytes)



from __future__ import annotations

from typing import Any, cast

try:
    import xgboost  # type: ignore[import]
except ModuleNotFoundError:
    pass  # error message in typing.py

import sarus_data_spec.typing as st

from .utils import pandas_or_value


async def xgb_classifier(  # to be renamed ?
    scalar: st.Scalar, *args: Any, **kwargs: Any
) -> xgboost.XGBClassifier:
    (ds_model, X, y), _ = scalar.parents()
    model = await cast(st.Scalar, ds_model).async_value()
    assert isinstance(model, xgboost.XGBClassifier)
    X = await pandas_or_value(X)
    y = await pandas_or_value(y)
    fitted_model = model.fit(X, y, *args, **kwargs)
    return fitted_model


async def xgb_classifier_predict(
    scalar: st.Scalar,
    *args: Any,
    **kwargs: Any,
) -> Any:
    (ds_model, X), _ = scalar.parents()
    model = await cast(st.Scalar, ds_model).async_value()
    assert isinstance(model, xgboost.XGBClassifier)
    X = await pandas_or_value(X)
    y_pred = model.predict(X, *args, **kwargs)
    return y_pred



from typing import Callable, MutableMapping, TypeVar

from sarus_data_spec.protobuf import type_name
from sarus_data_spec.protobuf.typing import Protobuf
import sarus_data_spec.typing as st


class Factory(st.Factory):
    """Can produce objects from protobuf messages"""

    type_name_create: MutableMapping[
        str, Callable[[Protobuf], st.HasProtobuf]
    ] = {}

    def register(
        self, name: str, create: Callable[[Protobuf], st.HasProtobuf]
    ) -> None:
        self.type_name_create[name] = create

    M = TypeVar('M', bound=Protobuf)

    def create(self, message: M) -> st.HasProtobuf[M]:
        return self.type_name_create[type_name(message)](message)



from __future__ import annotations

from typing import (
    Any,
    AsyncIterator,
    Callable,
    Collection,
    Dict,
    Iterator,
    List,
    Optional,
    Protocol,
    Tuple,
    runtime_checkable,
)

import pandas as pd
import pyarrow as pa

try:
    import tensorflow as tf
except ModuleNotFoundError:
    pass  # Warning is displayed by typing.py

from sarus_data_spec.storage.typing import HasStorage
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


@runtime_checkable
class Manager(st.Referrable[sp.Manager], HasStorage, Protocol):
    """Provide the dataset functionalities."""

    parquet_dir: str

    def schema(self, dataset: st.Dataset) -> st.Schema:
        ...

    def marginals(self, dataset: st.Dataset) -> st.Marginals:
        ...

    def to_arrow(
        self, dataset: st.Dataset, batch_size: int
    ) -> Iterator[pa.RecordBatch]:
        ...

    async def async_to_arrow(
        self, dataset: st.Dataset, batch_size: int
    ) -> AsyncIterator[pa.RecordBatch]:
        ...

    def to_parquet(self, dataset: st.Dataset) -> None:
        ...

    def to_pandas(self, dataset: st.Dataset) -> pd.DataFrame:
        ...

    async def async_to_pandas(self, dataset: st.Dataset) -> pd.DataFrame:
        ...

    def to_tensorflow(self, dataset: st.Dataset) -> tf.data.Dataset:
        ...

    async def async_to_tensorflow(
        self, dataset: st.Dataset
    ) -> tf.data.Dataset:
        ...

    def status(self, dataset: st.DataSpec) -> st.Status:
        """Reference to use to refer to this object."""
        ...

    def size(self, dataset: st.Dataset) -> st.Size:
        ...

    def bounds(self, dataset: st.Dataset) -> st.Bounds:
        ...

    def is_compliant(
        self,
        dataspec: st.DataSpec,
        kind: st.ConstraintKind,
        public_context: List[str],
        epsilon: Optional[float],
    ) -> bool:
        ...

    def variant(
        self,
        dataspec: st.DataSpec,
        kind: st.ConstraintKind,
        public_context: List[str],
        epsilon: Optional[float],
    ) -> Optional[st.DataSpec]:
        ...

    def variants(self, dataspec: st.DataSpec) -> Collection[st.DataSpec]:
        ...

    def variant_constraint(
        self, dataspec: st.DataSpec
    ) -> Optional[st.VariantConstraint]:
        ...

    def set_remote(self, dataspec: st.DataSpec) -> None:
        """Add an Attribute to tag the DataSpec as remotely fetched."""
        ...

    def is_remote(self, dataspec: st.DataSpec) -> bool:
        """Is the dataspec a remotely defined dataset."""
        ...

    def value(self, scalar: st.Scalar) -> st.DataSpecValue:
        ...

    async def async_value(self, scalar: st.Scalar) -> st.DataSpecValue:
        ...

    def infer_output_type(
        self,
        transform: st.Transform,
        *arguments: st.DataSpec,
        **named_arguments: st.DataSpec,
    ) -> Tuple[str, Callable[[st.DataSpec], None]]:
        ...

    def verifies(
        self,
        variant_constraint: st.VariantConstraint,
        kind: st.ConstraintKind,
        public_context: Collection[str],
        epsilon: Optional[float],
    ) -> bool:
        ...

    def foreign_keys(self, dataset: st.Dataset) -> Dict[st.Path, st.Path]:
        ...

    def primary_keys(self, dataset: st.Dataset) -> List[st.Path]:
        ...

    def sql(self, dataset: st.Dataset, query: str) -> List[Dict[str, Any]]:
        ...


@runtime_checkable
class HasManager(Protocol):
    """Has a manager."""

    def manager(self) -> Manager:
        """Return a manager (usually a singleton)."""
        ...



from types import TracebackType
import typing as t

from sarus_data_spec.context.state import (
    pop_global_context,
    push_global_context,
)
from sarus_data_spec.context.typing import Context
from sarus_data_spec.manager.typing import Manager
from sarus_data_spec.storage.typing import Storage
import sarus_data_spec.typing as st


class Public(Context):
    """A public context base"""

    def __init__(self) -> None:
        pass

    def factory(self) -> st.Factory:
        raise NotImplementedError

    def storage(self) -> Storage:
        raise NotImplementedError()

    def manager(self) -> Manager:
        raise NotImplementedError()

    def __enter__(self) -> Context:
        push_global_context(self)
        return self

    def __exit__(
        self,
        type: t.Optional[t.Type[BaseException]],
        value: t.Optional[BaseException],
        traceback: t.Optional[TracebackType],
    ) -> None:
        pop_global_context()
        return None



from typing import Collection, Dict, List, Optional, Tuple, cast

from sarus_data_spec.attribute import attach_properties
from sarus_data_spec.dataset import transformed
from sarus_data_spec.query_manager.typing import QueryManager
from sarus_data_spec.transform import synthetic
from sarus_data_spec.variant_constraint import variant_constraint
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st

ArgStruct = Tuple[List[int], List[str]]


def flatten_args(
    args: List[st.DataSpec], kwargs: Dict[str, st.DataSpec]
) -> Tuple[List[st.Dataset], List[st.Scalar], ArgStruct]:
    """Split args and kwargs into Datasets and Scalars."""
    flat_args = args + list(kwargs.values())
    ds_args = [
        cast(st.Dataset, arg)
        for arg in flat_args
        if arg.prototype() == sp.Dataset
    ]
    sc_args = [
        cast(st.Scalar, arg)
        for arg in flat_args
        if arg.prototype() == sp.Scalar
    ]

    ds_idx = [
        i for i, arg in enumerate(flat_args) if arg.prototype() == sp.Dataset
    ]
    sc_idx = [
        i for i, arg in enumerate(flat_args) if arg.prototype() == sp.Scalar
    ]
    idx = ds_idx + sc_idx
    struct = (idx, list(kwargs.keys()))

    return ds_args, sc_args, struct


def nest_args(
    ds_args: List[st.DataSpec],
    sc_args: List[st.DataSpec],
    struct: ArgStruct,
) -> Tuple[List[st.DataSpec], Dict[str, st.DataSpec]]:
    """Nest Datasets and Scalars into args and kwargs."""
    idx, keys = struct
    all_args = ds_args + sc_args
    flat_args = [all_args[idx.index(i)] for i in range(len(idx))]
    n_args = len(flat_args) - len(keys)
    args = flat_args[:n_args]
    kwargs = {key: val for key, val in zip(keys, flat_args[n_args:])}
    return args, kwargs


def attach_variant(
    original: st.DataSpec,
    variant: st.DataSpec,
    kind: st.ConstraintKind,
) -> None:
    sp_kind = sp.ConstraintKind.Value(kind.name)
    attach_properties(
        original, properties={sp.ConstraintKind.Name(sp_kind): variant.uuid()}
    )


def verifies(
    query_manager: QueryManager,
    variant_constraint: st.VariantConstraint,
    kind: st.ConstraintKind,
    public_context: Collection[str],
    epsilon: Optional[float],
) -> bool:
    if kind == st.ConstraintKind.PUBLIC:
        return verifies_public(variant_constraint=variant_constraint)

    elif kind == st.ConstraintKind.SYNTHETIC:
        return verifies_synthetic(variant_constraint=variant_constraint)

    elif kind == st.ConstraintKind.DP:
        raise NotImplementedError("DP verification")

    else:  # kind == st.ConstraintKind.PEP:
        raise NotImplementedError("PEP verification")


def verifies_public(variant_constraint: st.VariantConstraint) -> bool:
    return variant_constraint.constraint_kind() == st.ConstraintKind.PUBLIC


def verifies_synthetic(variant_constraint: st.VariantConstraint) -> bool:
    return variant_constraint.constraint_kind() in [
        st.ConstraintKind.PUBLIC,
        st.ConstraintKind.SYNTHETIC,
    ]


def compile(
    query_manager: QueryManager,
    dataspec: st.DataSpec,
    kind: st.ConstraintKind,
    public_context: List[str],
    epsilon: Optional[float],
) -> Optional[st.DataSpec]:
    """Returns a compliant Node or None."""

    if kind == st.ConstraintKind.SYNTHETIC:
        variant, _ = compile_synthetic(
            query_manager,
            dataspec,
            public_context,
        )
        return variant

    if not epsilon:
        raise ValueError("epsilon must be defined for PEP or DP to compile.")

    if kind == st.ConstraintKind.DP:
        raise NotImplementedError("DP compilation")

    elif kind == st.ConstraintKind.PEP:
        raise NotImplementedError("PEP compilation")

    else:
        raise ValueError(
            f"Privacy policy {kind} compilation " "not implemented yet"
        )


def compile_synthetic(
    query_manager: QueryManager,
    dataspec: st.DataSpec,
    public_context: Collection[str],
) -> Tuple[st.DataSpec, Collection[str]]:
    # Caching mechanisms
    if dataspec.verifies(
        kind=st.ConstraintKind.SYNTHETIC,
        public_context=public_context,
        epsilon=None,
    ):
        return dataspec, public_context

    for variant in dataspec.variants():
        if variant.verifies(
            kind=st.ConstraintKind.SYNTHETIC,
            public_context=public_context,
            epsilon=None,
        ):
            return variant, public_context

    if dataspec.prototype() == sp.Dataset:
        dataset = cast(st.Dataset, dataspec)
        # TODO remove this constraint once we have the whole
        # graph for a remote DataSpec
        if not dataspec.is_remote() and dataset.is_synthetic():
            variant_constraint(
                constraint_kind=st.ConstraintKind.SYNTHETIC,
                dataspec=dataspec,
            )
            return dataset, list(public_context)

    if not dataspec.is_remote() and dataspec.is_transformed():
        transform = dataspec.transform()
        args, kwargs = dataspec.parents()
        ds_args, sc_args, struct = flatten_args(args, kwargs)
        ds_syn_args_context = [
            compile_synthetic(query_manager, parent, public_context)
            for parent in ds_args
        ]
        sc_syn_args_context = [
            compile_synthetic(query_manager, parent, public_context)
            for parent in sc_args
        ]

        if len(ds_syn_args_context) > 0:
            ds_syn_args, ds_contexts = zip(*ds_syn_args_context)
        else:
            ds_syn_args, ds_contexts = [], ([],)
        if len(sc_syn_args_context) > 0:
            sc_syn_args, sc_contexts = zip(*sc_syn_args_context)
        else:
            sc_syn_args, sc_contexts = [], ([],)
        new_context = list(set(sum(ds_contexts + sc_contexts, [])))
        args, kwargs = nest_args(
            cast(List[st.DataSpec], list(ds_syn_args)),
            cast(List[st.DataSpec], list(sc_syn_args)),
            struct,
        )
        syn_variant = cast(
            st.DataSpec,
            transformed(
                transform,
                *args,
                dataspec_type=sp.type_name(dataspec.prototype()),
                dataspec_name=None,
                **kwargs,
            ),
        )
        variant_constraint(
            constraint_kind=st.ConstraintKind.SYNTHETIC,
            dataspec=syn_variant,
        )
        attach_variant(dataspec, syn_variant, kind=st.ConstraintKind.SYNTHETIC)
        return syn_variant, new_context

    # Source (non transformed) dataset
    if dataspec.prototype() == sp.Dataset:
        dataset = cast(st.Dataset, dataspec)
        syn_dataset = cast(st.Dataset, synthetic()(dataset))
        variant_constraint(
            constraint_kind=st.ConstraintKind.SYNTHETIC,
            dataspec=syn_dataset,
        )
        attach_variant(dataset, syn_dataset, kind=st.ConstraintKind.SYNTHETIC)
        new_context = list(public_context) + [syn_dataset.uuid()]
        return syn_dataset, new_context

    elif dataspec.prototype() == sp.Scalar:
        scalar = cast(st.Scalar, dataspec)
        assert scalar.is_model()
        new_context = list(public_context) + [scalar.uuid()]
        variant_constraint(
            constraint_kind=st.ConstraintKind.PUBLIC,
            dataspec=scalar,
        )
        return scalar, new_context

    raise ValueError("Uncaught compilation case.")


def is_public(dataspec: st.DataSpec) -> bool:
    if dataspec.prototype() == sp.Scalar:
        scalar = cast(st.Scalar, dataspec)
        if scalar.is_model():
            return True

    # Returns true if the DataSpec derives only from public
    if dataspec.is_transformed():
        args_parents, kwargs_parents = dataspec.parents()
        return all(
            [is_public(ds) for ds in args_parents]
            + [is_public(ds) for ds in kwargs_parents.values()]
        )

    # Default case
    return False



from itertools import count
from typing import Any, Generic, Tuple, TypeVar, cast
import asyncio
import logging

from sarus_data_spec.constants import CACHE, CACHE_PATH
from sarus_data_spec.context import global_context
from sarus_data_spec.status import (
    error,
    last_status,
    last_statuses,
    processing,
    ready,
)
import sarus_data_spec.protobuf as sp
import sarus_data_spec.typing as st


class DataSpecErrorStatus(Exception):
    ...


T = TypeVar("T")

logger = logging.getLogger(__name__)


class StatusAwareComputation(Generic[T]):
    """Singleton visitor class template to perform basic status checks.

    This class' purpose if to manage statuses, before, during and after a
    DataSpec's computation. It also manages caching.

    This class is intended to be subclassed and used in a Manager's
    `async_to_XXX` method. For instance, to implement `Manager.async_to_arrow`
    the idea is to implement a subclass ArrowComputation(Iterator[pa.
    RecordBatch]) and to do ArrowComputation.accept(dataspec).

    Subclasses must only implement the __read_from_cache__, __write_to_cache__
    and __process__ methods.
    """

    @classmethod
    async def accept(
        cls, dataspec: st.DataSpec, *args: Any, **kwargs: Any
    ) -> T:
        """Apply the appropriate method depending on the status."""

        # this is to fix current different status behavior between DOD
        # and DATASPEC
        hot_fix_status(dataspec)

        statuses = last_statuses(dataspec, task=CACHE)
        # a dataspec is always cached in one place, so we can take
        # the first status for that for now

        if len(statuses) == 0:
            return await cls.process(dataspec, *args, **kwargs)
        else:
            cache_task = cast(st.Stage, statuses[0].task(CACHE))
            if cache_task.ready():
                # TODO: cls.ready should take a stage rather than
                # the status
                return await cls.ready(dataspec, statuses[0])
            elif cache_task.pending():
                return await cls.pending(dataspec, *args, **kwargs)
            elif cache_task.processing():
                return await cls.processing(
                    dataspec, *args, interval=20, attempts=10, **kwargs
                )
            elif cache_task.error():
                return await cls.error(dataspec, *args, **kwargs)
            else:
                raise ValueError(f"Inconsistent status {statuses[0]}")

    @classmethod
    async def processing(
        cls,
        dataspec: st.DataSpec,
        *args: Any,
        attempts: int = 10,
        interval: int = 20,
        **kwargs: Any,
    ) -> T:
        """If processing, wait for the dataspec to be ready.

        Such a case can happen if another manager has launched the computation
        of this DataSpec's value. A timeout period is defined after which the
        computation is relaunched (this is a default behavior intended to be
        improved in the future).

        Args:
            attempts (int): number of attempts before timing out.
            interval (int): number of seconds between attempts.
        """
        attempt = count()
        while next(attempt) <= attempts:
            # TODO: in principle, here only one manager can be
            # processing the operation but have to be sure
            status = last_status(dataspec, task=CACHE)
            assert status
            if not status.processing():
                break
            await asyncio.sleep(interval)

        status = last_status(dataspec)
        assert status
        if status.processing():
            # Still processing, process anyway
            # TODO might not be a good idea to relaunch
            # a really long computation
            return await cls.process(dataspec, *args, **kwargs)
        else:
            # computation finished, revisit to define what to do
            return await cls.accept(dataspec, *args, **kwargs)

    @classmethod
    async def error(
        cls,
        dataspec: st.DataSpec,
        *args: Any,
        **kwargs: Any,
    ) -> T:
        """The DataSpec already has an Error status.
        In this case we clear the error and relaunch
        the process"""
        # we want to erase all the statuses of the
        # dataspec concerning the CACHE
        statuses = dataspec.storage().referring(
            dataspec, type_name=sp.type_name(sp.Status)
        )
        for status in statuses:
            status = cast(st.Status, status)
            if status.task(CACHE) is not None:
                dataspec.storage().delete(status.uuid())
        return await cls.accept(dataspec, *args, **kwargs)

    @classmethod
    async def pending(
        cls, dataspec: st.DataSpec, *args: Any, **kwargs: Any
    ) -> T:
        """The DataSpec has been added to the computation queue.

        No manager has undertaken the computation yet."""
        return await cls.process(dataspec, *args, **kwargs)

    @classmethod
    async def process(
        cls, dataspec: st.DataSpec, *args: Any, **kwargs: Any
    ) -> T:
        """Compute the DataSpec's value.

        Set the DataSpec's status to PROCESSING and launch the computation
        defined in the subclass.

        If an error occurs, set the status to ERROR with the Exception
        representation.

        If no error occurs, write the result to cache, set the status to READY
        with the cache path and return the actual value.
        """
        manager = global_context().manager()
        processing(
            dataspec=dataspec,
            manager=manager,
            task=CACHE,
        )
        try:
            result = await cls.__process__(dataspec, *args, **kwargs)
        except Exception as e:
            error(
                dataspec=dataspec,
                manager=manager,
                task=CACHE,
                properties={"message": repr(e)},
            )
            logger.exception(e)
            raise DataSpecErrorStatus(repr(e))
        else:
            cache_path, result = await cls.__write_to_cache__(dataspec, result)
            ready(
                dataspec=dataspec,
                manager=manager,
                task=CACHE,
                properties={CACHE_PATH: cache_path},
            )
            return result

    @classmethod
    def cache_path(cls, status: st.Status) -> str:
        assert status.ready()
        task_cache = status.task(CACHE)
        assert task_cache is not None
        path = task_cache.properties()[CACHE_PATH]
        if path:
            return path
        raise ValueError('Cache was not found')

    @classmethod
    async def ready(cls, dataspec: st.DataSpec, status: st.Status) -> T:
        """Read the value from the cache."""
        return await cls.__read_from_cache__(dataspec, status)

    @classmethod
    async def __read_from_cache__(
        cls, dataspec: st.DataSpec, status: st.Status
    ) -> T:
        # Should be implemented for each to_XXX method
        raise NotImplementedError

    @classmethod
    async def __write_to_cache__(
        cls, dataspec: st.DataSpec, data: T
    ) -> Tuple[str, T]:
        # Should be implemented for each to_XXX method
        raise NotImplementedError

    @classmethod
    async def __process__(
        cls,
        dataspec: st.DataSpec,
        *args: Any,
        **kwargs: Any,
    ) -> T:
        # Must be reimplemented for each to_XXX method
        raise NotImplementedError


def hot_fix_status(dataspec: st.DataSpec) -> None:
    """Currently, in DOD, status tasks are given
    by the protobuf name, so we create a status for
    the cache task manually
    """
    statuses = last_statuses(dataspec)
    for status in statuses:
        for task_name in status.protobuf().task_stages.keys():
            path = status.task(task_name).properties()["cache"]  # type:ignore
            if path:
                logging.warn(
                    f'Cache was found in task {task_name},'
                    f' but it should be in {CACHE}, added by hot_fix'
                )
                ready(dataspec, task=CACHE, properties={CACHE_PATH: path})



from collections import defaultdict
from itertools import chain, combinations
from typing import (
    Collection,
    DefaultDict,
    Final,
    FrozenSet,
    List,
    MutableMapping,
    Optional,
    Set,
    Union,
    cast,
)

from sarus_data_spec.protobuf.typing import (
    ProtobufWithUUID,
    ProtobufWithUUIDAndDatetime,
)
from sarus_data_spec.typing import Referrable, Referring

SEP: Final[str] = ','


def referrable_collection_string(
    values: Collection[Referrable[ProtobufWithUUID]],
) -> str:
    return SEP.join(sorted(value.uuid() for value in values))


def referrable_collection_set(
    values: Collection[Referrable[ProtobufWithUUID]],
) -> FrozenSet[str]:
    return frozenset(value.uuid() for value in values)


class Storage:
    """Simple local Storage."""

    def __init__(self) -> None:
        # A Store to save (timestamp, type_name, data, relating data)
        self._referrables: MutableMapping[
            str, Referrable[ProtobufWithUUID]
        ] = dict()
        self._referring: DefaultDict[str, Set[str]] = defaultdict(set)

    def store(self, value: Referrable[ProtobufWithUUID]) -> None:
        # Checks the value for consistency
        assert value._frozen()
        self._referrables[value.uuid()] = value

        if isinstance(value, Referring):
            value = cast(Referring[ProtobufWithUUID], value)
            referred_combinations = chain.from_iterable(
                combinations(value.referred(), r) for r in range(1, 3)
            )
            for combination in referred_combinations:
                self._referring[referrable_collection_string(combination)].add(
                    value.uuid()
                )

    def referrable(self, uuid: str) -> Optional[Referrable[ProtobufWithUUID]]:
        return self._referrables.get(uuid, None)

    def referring(
        self,
        referred: Union[
            Referrable[ProtobufWithUUID],
            Collection[Referrable[ProtobufWithUUID]],
        ],
        type_name: Optional[str] = None,
    ) -> Collection[Referring[ProtobufWithUUID]]:
        if isinstance(referred, Referrable):
            referring_uuids = self._referring[referred.uuid()]
        else:
            referring_uuids = self._referring[
                referrable_collection_string(referred)
            ]
        referrings = [self.referrable(uuid) for uuid in referring_uuids]
        if type_name is not None:
            referrings = [
                item
                for item in referrings
                if (item is not None) and (item.type_name() == type_name)
            ]
        return referrings  # type:ignore

    def last_referring(
        self,
        referred: Union[
            Referrable[ProtobufWithUUID],
            Collection[Referrable[ProtobufWithUUID]],
        ],
        type_name: str,
    ) -> Optional[Referring[ProtobufWithUUIDAndDatetime]]:
        """Last value referring to one referred.
        This implementation is not very efficient"""
        referrings = cast(
            Collection[Referring[ProtobufWithUUIDAndDatetime]],
            self.referring(referred, type_name),
        )
        if len(referrings) > 0:
            return max(referrings, key=lambda r: r.protobuf().datetime)
        else:
            return None

    def type_name(
        self, type_name: str
    ) -> Collection[Referrable[ProtobufWithUUID]]:
        return {
            ref
            for ref in self._referrables.values()
            if ref.type_name() == type_name
        }

    def all_referrings(self, uuid: str) -> List[str]:
        """Returns a list all items referring to a Referrable recursively."""
        target = self.referrable(uuid)

        to_delete, to_check = set(), {target}
        while len(to_check) > 0:
            node = to_check.pop()
            if not node:
                continue
            to_delete.add(node)
            deps = node.referring()
            if not deps:
                continue
            for dep in deps:
                if dep not in to_delete:
                    to_check.add(dep)

        return [msg.uuid() for msg in to_delete]

    def delete(self, uuid: str) -> None:
        """Delete a Referrable and all elements referring to it to let the
        storage in a consistent state."""
        uuids_to_delete = set(self.all_referrings(uuid))

        self._referrables = {
            k: v
            for k, v in self._referrables.items()
            if k not in uuids_to_delete
        }

        self._referring = defaultdict(
            set,
            {
                k: v - uuids_to_delete
                for k, v in self._referring.items()
                if k not in uuids_to_delete
            },
        )
